from release_mgr.git import get_contributors_between, get_messages_between


def build_changelog(messages):
    return "\n".join(
        [
            f"- {message}"
            for message in messages
            if message
            and "release: " not in message
            and "Merge pull request" not in message
        ],
    )


def build_contributors(emails):
    return "\n".join({f"- {email}" for email in emails if email})


def build_release_notes(
    version,
    messages,
    emails,
    title="",
):
    if title:
        release_notes = f"# {title}\n\n"
    else:
        release_notes = f"# Release {version}\n\n"

    release_notes += build_changelog(messages)
    release_notes += "\n\n# Contributors to this Release\n\n"
    release_notes += build_contributors(emails)
    release_notes += "\n"

    return release_notes


def release_notes_for_version(
    version,
    version_commit,
    last_version_commit,
):
    messages = get_messages_between(last_version_commit, version_commit)
    emails = get_contributors_between(last_version_commit, version_commit)
    return build_release_notes(
        version=version,
        messages=messages,
        emails=emails,
    )



import requests


def create_release(
    token,
    repo,
    tag_name,
    name,
    body,
    draft=False,
    prerelease=False,
):
    headers = {
        "Authorization": f"token {token}",
        "Accept": "application/vnd.github.v3+json",
    }
    body = {
        "tag_name": tag_name,
        "name": name,
        "body": body,
        "draft": draft,
        "prerelease": prerelease,
    }

    response = requests.post(
        f"https://api.github.com/repos/{repo}/releases",
        headers=headers,
        json=body,
    )
    response.raise_for_status()



import re

from release_mgr.version_files.version_file import VersionFile


class PyProjectTOML(VersionFile):
    filename = "pyproject.toml"
    pattern = re.compile(r'(version = )"[0-9]*.[0-9]*.[0-9]*"')
    replace_pattern = r'\1"{new_version}"'



import re

from release_mgr.version_files.version_file import VersionFile


class CargoTOML(VersionFile):
    filename = "Cargo.toml"
    pattern = re.compile(r'^version = "[0-9]*\.[0-9]*\.[0-9]*"$', re.MULTILINE)
    replace_pattern = 'version = "{new_version}"'



from release_mgr.git import git
from release_mgr.version_files.cargo_toml import CargoTOML
from release_mgr.version_files.dunder_version import DunderVersion
from release_mgr.version_files.lerna_json import LernaJSON
from release_mgr.version_files.package_json import PackageJSON
from release_mgr.version_files.pyproject import PyProjectTOML
from release_mgr.version_files.setup_py import SetupPy

VERSION_FILES = [
    PackageJSON,
    LernaJSON,
    SetupPy,
    PyProjectTOML,
    DunderVersion,
    CargoTOML,
]


def update_version_files(version):
    updated = update_vfs(version)
    if updated:
        git("commit", "-m", f"release: {version}")
        git("push")


def update_vfs(version):
    version_file_updated = False
    for version_file in VERSION_FILES:
        vf = version_file.from_file()
        if not vf:
            continue

        updated = vf.update(str(version))
        if updated:
            print("Updated version in:", vf.filename)

        version_file_updated = version_file_updated or updated

    return version_file_updated



import re

from release_mgr.git import get_commit_for_tag, get_tags, git


class Version:
    """A semver object."""

    release_ver = re.compile(r"^(v)?[0-9]+\.[0-9]+\.[0-9]+(-[A-z0-9]+)?$")

    def __init__(self, major, minor, patch, suffix=None, use_v_prefix=False):
        self.major = major
        self.minor = minor
        self.patch = patch
        self.suffix = suffix
        self.use_v_prefix = use_v_prefix

    @staticmethod
    def is_version_string(verstr):
        return bool(Version.release_ver.match(verstr))

    @classmethod
    def latest_version(cls):
        tags = [
            tag[len("refs/tags/") :] for tag in get_tags() if tag[len("refs/tags/") :]
        ]
        if not tags:
            commits = git("log", "--reverse", "--format=%H").split("\n")
            return (
                commits[0],
                cls(0, 0, 0),
            )

        versions = sorted(map(Version.from_str, tags))
        latest_version = versions[-1]
        last_version_commit = get_commit_for_tag(str(latest_version))
        return (
            last_version_commit,
            latest_version,
        )

    @classmethod
    def from_str(cls, verstr):
        if not Version.is_version_string(verstr):
            raise Exception("Got unexpected input: {verstr}".format(verstr=verstr))

        major, minor, patch = verstr.split(".")
        suffix = None
        if "-" in patch:
            patch, suffix = patch.split("-")

        use_v_prefix = False
        if major.startswith("v"):
            major = major[1:]
            use_v_prefix = True

        return cls(
            int(major),
            int(minor),
            int(patch),
            suffix,
            use_v_prefix=use_v_prefix,
        )

    def increment_patch(self):
        self.patch += 1

    def increment_minor(self):
        self.minor += 1
        self.patch = 0

    def increment_major(self):
        self.major += 1
        self.minor = 0
        self.patch = 0

    def __str__(self):
        return "{prefix}{major}.{minor}.{patch}{suffix}".format(
            major=self.major,
            minor=self.minor,
            patch=self.patch,
            suffix="-" + self.suffix if self.suffix else "",
            prefix="v" if self.use_v_prefix else "",
        )

    def __eq__(self, other):
        if not isinstance(other, Version):
            return False

        return (
            self.major == other.major
            and self.minor == other.minor
            and self.patch == other.patch
            and self.suffix == other.suffix
        )

    def __lt__(self, other):
        for version_part in ["major", "minor", "patch"]:
            ours = getattr(self, version_part)
            theirs = getattr(other, version_part)
            if ours > theirs:
                return False

            if theirs > ours:
                return True

            # Same version part value

        if self.suffix == other.suffix:
            return False

        suffix_precedence = {
            "beta": -1,
            "tc": 0,
            "rc": 1,
            None: 2,
        }

        our_suffix = suffix_precedence.get(self.suffix, -2)
        their_suffix = suffix_precedence.get(other.suffix, -2)
        return our_suffix < their_suffix

    def __le__(self, other):
        return self == other or self < other



from subprocess import check_output


def git(*args):
    cmd = ["git"]
    cmd.extend(args)
    return check_output(cmd).decode("utf-8")


def get_repo():
    remotes = check_output(["git", "remote", "-v"]).decode("utf-8").split("\n")
    for remote in remotes:
        if not remote:
            continue

        first_segment, _ = remote.split(" ")
        _, url = first_segment.split("\t")

        if "github.com" not in url:
            continue

        if url.startswith("ssh") or url.startswith("git@"):
            repo = url.split(":")[-1]
            if repo.endswith(".git"):
                repo = repo[: -len(".git")]

            return repo
        else:
            return "/".join(url.split("/")[-2:])


def get_tags():
    return (
        check_output(
            [
                "git",
                "for-each-ref",
                "--sort=-taggerdate",
                "--format",
                "%(refname)",
                "refs/tags",
            ],
        )
        .decode("utf-8")
        .split("\n")
    )


def get_commit_for_tag(tag):
    return (
        check_output(["git", "show", "--no-patch", "--pretty=%H", tag])
        .decode("utf-8")
        .strip()
    )


def get_contributors_between(oldest_commit, latest_commit):
    return (
        check_output(
            [
                "git",
                "log",
                "--pretty=format:%an <%ae>",
                f"{oldest_commit}...{latest_commit}",
            ],
        )
        .decode("utf-8")
        .split("\n")
    )


def get_messages_between(oldest_commit, latest_commit):
    return (
        check_output(
            [
                "git",
                "log",
                "--pretty=format:%h %s",
                f"{oldest_commit}...{latest_commit}",
            ],
        )
        .decode("utf-8")
        .split("\n")
    )



import pytest
from release_mgr.version import Version
from release_mgr.version_files.cargo_toml import CargoTOML

CARGO_TOML = """
[package]
authors = ["Mathew Robinson <chasinglogic@gmail.com>"]
name = "licensure"
version = ">>>VERSION<<<"
keywords = ["licensing", "cli", "tool", "license"]
license = "GPL-3.0"
readme = "README.md"
description = "A software license management tool"
repository = "https://github.com/chasinglogic/licensure"
homepage = "https://github.com/chasinglogic/licensure"
edition = "2018"

[badges]
travis-ci = { repository = "chasinglogic/licensure", branch = "master" }

[dependencies]
chrono = "0.4.2"
clap = "2.31.2"
regex = "1.0.0"
serde = { version = "1.0", features = ["derive"] }
serde_yaml = "0.8.21"
log = "0.4.8"
simplelog = "0.11.0"
reqwest = "0.9.19"
textwrap = "0.14.2"
"""


@pytest.mark.parametrize(
    ("old_version", "new_version"),
    [
        ("0.2.0", "0.2.1"),
        ("0.14.2", "0.14.3"),
    ],
)
def test_cargo_toml_update(old_version, new_version):
    cargo = CargoTOML(content=CARGO_TOML.replace(">>>VERSION<<<", old_version))
    new_content = cargo.subst(new_version)
    assert new_content == CARGO_TOML.replace(">>>VERSION<<<", new_version)



import os
import re

from release_mgr.git import git


class VersionFile:
    """
    Represents a file which would contain the version.
    """

    filename = ""
    pattern = re.compile("")
    replace_pattern = "{new_version}"

    def __init__(self, content=""):
        self.content = content

    @classmethod
    def from_file(cls):
        if os.path.exists(cls.filename):
            obj = cls()
            obj.load_content()
            return obj

        return None

    def load_content(self):
        with open(self.filename, encoding="utf-8") as existing:
            self.content = existing.read()

    def subst(self, version):
        return self.pattern.sub(
            self.replace_pattern.format(new_version=version),
            self.content,
        )

    def update(self, version: str):
        if not self.pattern:
            raise Exception("Tried to update a file without a pattern!")

        if not self.content:
            self.load_content()

        if self.pattern.search(self.content) is None:
            return False

        new_content = self.subst(version)

        with open(self.filename, "w", encoding="utf-8") as new:
            new.write(new_content)

        git("add", self.filename)
        return True



import re

from release_mgr.version_files.version_file import VersionFile


class SetupPy(VersionFile):
    filename = "setup.py"
    pattern = re.compile(r'(version=)[\'"][0-9]*.[0-9]*.[0-9]*[\'"]')
    replace_pattern = r'\1"{new_version}"'



"""A task management tool that integrates with 3rd party services."""

import re
from os import path

from setuptools import find_packages, setup

with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

init_py = path.join(path.dirname(__file__), "release_mgr", "__init__.py")
with open(init_py, encoding="utf-8") as f:
    init_content = f.read()
    version = re.search(r"__version__ = ['\"]([^'\"]+)['\"]", init_content, re.M).group(
        1
    )

setup(
    name="release-mgr",
    version=version,
    url="https://github.com/chasinglogic/gh-release",
    license="GPL-3.0",
    author="Mathew Robinson",
    author_email="chasinglogic@gmail.com",
    description="A simple tool for managing software releases on GitHub.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    platforms="any",
    install_requires=requirements,
    entry_points={"console_scripts": ["release-mgr = release_mgr.cli:main"]},
    classifiers=[
        # As from https://pypi.org/classifiers/
        "Development Status :: 4 - Beta",
        # 'Development Status :: 5 - Production/Stable',
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: POSIX",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import pytest

from release_mgr.version import Version


@pytest.mark.parametrize(
    ("verstring", "expected"),
    [
        ("0.0.0", True),
        ("0.0.0-beta", True),
        ("0.1.0", True),
        ("1010.100.0", True),
        ("1010.100.0-suffix", True),
        ("mat", False),
    ],
)
def test_is_version_string(verstring, expected):
    assert Version.is_version_string(verstring) == expected


def test_sorted():
    expected = [
        Version(0, 1, 0),
        Version(0, 2, 0),
        Version(0, 2, 1),
        Version(0, 3, 0),
        Version(0, 3, 1, suffix="beta"),
        Version(0, 3, 1),
        Version(1, 0, 0),
    ]

    inputs = [
        "0.1.0",
        "0.3.1",
        "0.2.1",
        "0.3.1-beta",
        "0.2.0",
        "0.3.0",
        "1.0.0",
    ]

    versions = [Version.from_str(verstr) for verstr in inputs]

    assert list(sorted(versions)) == expected



import glob
import re

from release_mgr.version_files.version_file import VersionFile


class DunderVersion(VersionFile):
    pattern = re.compile(r'__version__ = ["\'][0-9]*.[0-9]*.[0-9]*["\']')
    replace_pattern = '__version__ = "{new_version}"'

    @classmethod
    def from_file(cls):
        for pattern in [
            "src/*/version.py",
            "*/version.py",
            "src/*/__init__.py",
            "*/__init__.py",
        ]:
            files = glob.glob(pattern)
            if not files:
                continue

            files = [f for f in files if not f.startswith("test")]

            obj = cls()
            obj.filename = files[0]
            obj.load_content()
            if "__version__" not in obj.content:
                continue

            return obj

        return None



import json

from release_mgr.git import git
from release_mgr.version_files.version_file import VersionFile


class PackageJSON(VersionFile):
    filename = "package.json"

    def update(self, version: str):
        with open("package.json") as pkg_json_file:
            pkg_json = json.loads(pkg_json_file.read())
            pkg_json["version"] = version

        with open("package.json", "w") as pkg_json_file:
            json.dump(
                pkg_json,
                pkg_json_file,
                indent=2,
            )

        git("add", self.filename)
        return True



#!/usr/bin/env python3

import os
import sys
from copy import deepcopy

import click

from release_mgr.git import get_commit_for_tag, get_repo, git
from release_mgr.github import create_release
from release_mgr.release import release_notes_for_version
from release_mgr.version import Version
from release_mgr.version_files import update_version_files


@click.command()
@click.option(
    "--version",
    "-v",
    type=str,
    default="",
    help="Specify the version to go to",
)
@click.option(
    "--pre-release",
    "-l",
    is_flag=True,
    help="Indicates this is a pre-release",
)
@click.option(
    "--draft",
    "-d",
    is_flag=True,
    help="Indicates this is a draft release",
)
@click.option(
    "--repo",
    "-r",
    type=str,
    default=None,
    help="The repository to create the release on in :owner/:repo format, "
    "will attempt to parse from git remotes if not given",
)
@click.option(
    "--title",
    "-t",
    is_flag=True,
    help="If given use the release name as the markdown title, otherwise title "
    "is omitted for Github style formatting",
)
@click.option(
    "--minor",
    "-m",
    is_flag=True,
    help="Bump minor version",
)
@click.option(
    "--major",
    "-j",
    is_flag=True,
    help="Bump major version",
)
@click.option(
    "--patch",
    "-p",
    is_flag=True,
    help="Bump patch version",
)
@click.option(
    "--skip-version-files",
    "-s",
    is_flag=True,
    help="Don't try to update version metadata files (package.json, setup.py etc.)",
)
@click.option(
    "--skip-upload",
    is_flag=True,
    help="Don't try to create a release on github and don't push the commits",
)
def main(
    version,
    major,
    patch,
    minor,
    skip_version_files,
    pre_release,
    draft,
    repo,
    title,
    skip_upload,
):
    """
    A simple tool for managing software releases on GitHub.
    """
    if not any((patch, minor, major, version)):
        print("Must provide one of --version, --major, --minor, or --patch.")
        sys.exit(1)

    token = os.getenv("GITHUB_TOKEN")
    if not token:
        print("$GITHUB_TOKEN must be set.")
        sys.exit(1)

    last_version_commit, last_version = Version.latest_version()
    if version:
        version = Version.from_str(version)
    else:
        version = deepcopy(last_version)
        if patch:
            version.increment_patch()
        if minor:
            version.increment_minor()
        if major:
            version.increment_major()

    version_commit = get_commit_for_tag("HEAD")
    release_notes = release_notes_for_version(
        version,
        version_commit,
        last_version_commit,
    )

    if title:
        release_notes = f"# {version}\n\n{release_notes}"

    repo = repo or get_repo()

    print("Creating release", version, version_commit)
    print("Previous version", last_version, last_version_commit)
    print("Pre-release?", pre_release)
    print("Draft release?", draft)
    print("Repository", repo)
    print("============= Release Notes ============")
    print(release_notes)

    ans = input("Does this look correct? (y/N) ")
    if not ans.startswith("y"):
        return

    if not skip_version_files:
        update_version_files(version)

    git("tag", str(version))
    if not skip_upload:
        git("push", "--tags")

    try:
        if not skip_upload:
            create_release(
                token=token,
                repo=repo,
                tag_name=str(version),
                name=str(version),
                body=release_notes,
                draft=draft,
                prerelease=pre_release,
            )
    except Exception as exc:
        print("Failed to create release!")
        print(exc)


if __name__ == "__main__":
    main()
